// BuildDirectory
// Given a file or directory who's parent is a valid directory in the
// filesystem, add it to the parent's .jfs file. If a directory was given,
// recursively traverse all the subdirectories doing the same thing.
//
// Files and directories encountered that need to be added to a .jfs file
// are marked as being owned by the owner of the parent directory. Files
// encounterd that do not need to be added cause the .jfs file to be
// unchanged.
import java.io.*;
import java.net.URLConnection;

public class BuildDirectory
{
	public static void main(String argv[])
	{
	if (argv.length !=  2)
		die("usage: java BuildDirectory <JFS root> <directory>|<file>");
	FileSystem.init(argv[0]);
	JFSserver.loadcache();

	// Get the parent directory, if there is one
	String par = null;
	try { par = FileSystem.parent(argv[1]); }
	catch(BadPathException e) {
		die("Bogus pathname : "+argv[1]);
		}
	JFSdirectory pardir = null;
	JFSfile parfile = null;
	if (par.length() == 0) {
		// This is the root directory. Only need to call updatedir()
		updatedir(argv[0], "/", null, null);
		}
	else {
		// Some other file or directory
		try { pardir = FileSystem.getdir(par); }
		catch(BadPathException e) {
			die("The directory "+par+" is not in the filesystem");
			}
		try { parfile = FileSystem.statfile(par); }
		catch(BadPathException e) {
			die("Could not stat "+par+" : "+e.getMessage());
			}

		// Is the file given a file or directory?
		File argfile = new File(argv[0]+"/"+argv[1]);
		if (!argfile.exists())
			die("File "+argv[1]+" not found");
		int oldlen = pardir.files.size();
		if (argfile.isDirectory()) {
			// Recursively update
			updatedir(argv[0], argv[1], pardir, parfile);
			}
		else {
			// Just update one file
			updatefile(argv[0], argv[1], pardir, parfile, null);
			}
	
		// Write out directory, if needed
		if (pardir.files.size() != oldlen) {
			try { FileSystem.putdir(pardir, par); }
			catch(BadPathException e) {
				warn("Failed to write directory "+par);
				}
			}
		}
	}

	// updatedir
	// Add the directory to it's parent.
	// Recursively check all files and directories of this one, adding
	// them as well.
	static void updatedir(String root, String path,
			      JFSdirectory pardir, JFSfile parfile)
	{
	String base = path.substring(path.lastIndexOf('/')+1);

	JFSfile myinfo = null;
	if (path.equals("/")) {
		// / has no parent, and needs no updating
		try { myinfo = FileSystem.statfile("/"); }
		catch(BadPathException e) {
			die("Could not stat /");
			}
		}
	else {
		// Update this directory in it's parent
		updatefile(root, path, pardir, parfile, "jfs/directory");
		myinfo = pardir.find(base);
		}

	// This isn't a directory in the .jfs file
	if (!myinfo.type.equals("jfs/directory"))
		return;

	// Try to get the existing directory
	JFSdirectory mydir = null;
	int oldlen = -1;
	try {
		mydir = FileSystem.getdir(path);
		oldlen = mydir.files.size();
		}
	catch(BadPathException e) {
		// none.. so create an emtpy one
		mydir = new JFSdirectory();
		}

	// Add each local file or directory
	File myfile = new File(root+"/"+path);
	String files[] = myfile.list();
	for(int i=0; i<files.length; i++) {
		if (files[i].charAt(0) == '.' || files[i].indexOf(',') != -1)
			continue;	// .jfs or version file
		File addfile = new File(root+"/"+path+"/"+files[i]);
		if (addfile.isDirectory())
			updatedir(root, path+"/"+files[i], mydir, myinfo);
		else
			updatefile(root, path+"/"+files[i], mydir, myinfo,null);
		}

	// Write out this directory if needed
	if (mydir.files.size() != oldlen) {
		try { FileSystem.putdir(mydir, path); }
		catch(BadPathException e) {
			warn("Failed to write directory "+path);
			}
		}
	}

	// updatefile
	// Add the given file to the JFSdirectory structure of it's
	// parent. The new file inherits the ownership rights of
	// it's parent, and has the MIME type determined by it's extension
	// (if the given type is null).
	static void updatefile(String root, String path, JFSdirectory pardir,
			       JFSfile parfile, String type)
	{
	String base = path.substring(path.lastIndexOf('/')+1);

	// Does it already exist?
	if (pardir.find(base) != null)
		return;

	// Is it a useless temp file?
	if (base.endsWith(".bak") || base.endsWith(".swp"))
		return;

	// Work out mime type
	if (type == null) {
		type = URLConnectionHack.guessContentTypeFromName(base);
		if (type == null || type.equals("content/unknown"))
			type = "application/octec-stream";
		}

	// Create new file structure
	JFSfile nf = new JFSfile();
	nf.name = base;
	nf.type = type;
	nf.multiversion = false;
	JFSversion nv = new JFSversion();
	nv.number = 1;
	nv.creator = "root";
	nv.time = System.currentTimeMillis();
	nv.size = (int)(new File(root+"/"+path).length());
	nf.versions.addElement(nv);
	nf.users = parfile.users;
	pardir.files.addElement(nf);

	System.out.println("Added "+path+" ("+type+")");
	}

	// warn
	// Print a warning message
	static void warn(String msg)
	{
	System.err.println(msg);
	}

	// die
	// Print a message and exit
	static void die(String msg)
	{
	warn(msg);
	System.exit(1);
	}
}

// URLConnectionHack
// This class only exists to the handy guessContentTypeFromName() method
// in URLConnection can be accessed.
class URLConnectionHack extends URLConnection
{
	URLConnectionHack() { super(null); }
	public void connect() { }

	public static String guessContentTypeFromName(String file)
	{
	String map[][] = { { ".phtml", "text/x-phtml" },
			   { ".shtml", "text/x-shtml" },
			   { ".map", "text/x-imagemap" },
			   { ".cgi", "text/plain" },
			   { ".pl", "text/plain" },
			   { ".c", "text/plain" },
			   { ".java", "text/plain" },
			   { ".doc", "application/msword" },
			   { ".xls", "application/excel" },
			   { ".ppt", "application/powerpoint" },
			   { ".g3", "image/g3" },
			   { ".pdf", "application/pdf" },
			   { ".class", "application/octec-stream" } };
	for(int i=0; i<map.length; i++)
		if (file.endsWith(map[i][0]))
			return map[i][1];
	return URLConnection.guessContentTypeFromName(file);
	}
}

