// JFSShell
// A text-mode interface to a JFS server. This class is independant of the
// actual means of displaying the text (such as output to stdout or writing
// to a canvas). Provided commands are:
//  cd <directory>		- Change directory
//  pwd				- Display current directory
//  ls <file|directory>*	- List files or directories in short form
//  dir <file|directory>*	- List files or directories in long form
//  create <file> <type>	- Create an empty file of some type
//  rm <file|directory>+	- Delete files or directories
//  cp <file1> <file2>		- Copy one file to another
//  mv <file1> <file2>		- Move one file to another
//  mkdir <dir>			- Create a directory
//  open <file> [program]	- Open a file with it's handler 
//  run [program]		- Run the given class
//  help			- Display these commands
//  info <file>			- Display file information
//  uinfo <user>		- Display user details
//  ginfo <group>		- Display group details
import java.io.*;
import java.util.*;

public class JFSShell extends LineOutputStream
{
	LineOutputStream resp;
	StringBuffer line = new StringBuffer();	// current line
	JFSclient client;

	// Creates a new shell that will write command results to the given
	// stream. Commands are sent to the shell through the write() methods
	JFSShell(JFSclient c, LineOutputStream r)
	{
	client = c;
	resp = r;
	resp.puts("Connected to JFS server "+c.serverhost+" as "+c.authname);
	prompt();
	}

	// write
	// Add one byte to the command line, and if this is a \n execute the
	// command. It is assumed that the byte given has been echoed to the
	// display by the sender.
	public void write(int b)
	{
	char c = (char)b;
	if (c == '\n') {
		// Parse and execute
		StringTokenizer tok=new StringTokenizer(line.toString(), " \t");
		line.setLength(0);
		if (tok.countTokens() != 0) {
			String cmd = tok.nextToken();
			String par[] = new String[tok.countTokens()];
			for(int i=0; i<par.length; i++)
				par[i] = tok.nextToken();
			if (cmd.equals("ls")) ls(par);
			else if (cmd.equals("dir")) dir(par);
			else if (cmd.equals("pwd")) pwd(par);
			else if (cmd.equals("cd")) cd(par);
			else if (cmd.equals("info")) info(par);
			else if (cmd.equals("open")) open(par);
			else if (cmd.equals("run")) run(par);
			else if (cmd.equals("create")) create(par);
			else if (cmd.equals("rm")) rm(par);
			else {
				String tmp[] = { cmd };
				open(tmp);
				}
			}
		prompt();
		}
	else
		line.append(c);
	}

	// Write a series of bytes
	public void write(byte b[], int o, int l)
	{ for(int i=0; i<l; i++) write(b[i+o]); }
	public void write(byte b[])
	{ for(int i=0; i<b.length; i++) write(b[i]); }

	// prompt
	// Display a prompt
	void prompt()
	{
	try {
		resp.write((byte)'>');
		resp.write((byte)' ');
		resp.flush();
		}
	catch(IOException e) { }
	}

	// fullpath
	// Given a pathname relative to the current directory, expand it to
	// an abolute path and remove the ;version part (if any)
	String fullpath(String d)
	{
	int idx = d.lastIndexOf(';');
	if (idx != -1)
		d = d.substring(0, idx);
	String full = (d.charAt(0)=='/'||d.charAt(0)=='~') ?
			 d : (client.currentdir+"/"+d);
	StringTokenizer tok = new StringTokenizer(full, "/");
	String simple = "";
	while(tok.hasMoreTokens()) {
		String bit = tok.nextToken();
		if (bit.equals(".."))
			simple = simple.substring(0, simple.lastIndexOf('/'));
		else if (bit.equals("."))
			;	// ignore this part
		else if (bit.charAt(0) == '~') {
			// a home directory
			JFSuser u = null;
			try { u = client.uinfo(bit.substring(1)); }
			catch(RequestException e) { }
			if (u != null)
				simple += u.home;
			else
				simple += "/"+bit;
			}
		else
			simple += "/"+bit;
		}
	return simple.length()==0 ? "/" : simple;
	}

	// version
	// Given a filename of the form name;version, return the version
	// part (or 0)
	int version(String d)
	{
	int idx = d.lastIndexOf(';');
	if (idx != -1)
		return Integer.parseInt(d.substring(idx+1));
	return 0;
	}

	// pad
	// Make a string the given length, by padding with spaced or cropping
	String pad(String s, int n)
	{
	int d = n - s.length();
	if (d < 0) return s.substring(0, n);
	StringBuffer b = new StringBuffer();
	for(int i=0; i<d; i++) b.append(' ');
	return s+b;
	}

	// ls
	// Display a short-form directory listing
	void ls(String p[])
	{
	if (p.length == 0) {
		// List the current directory
		String tmp[] = { client.currentdir };
		ls(tmp);
		}
	else {
		// List the directories given
		for(int i=0; i<p.length; i++) {
			JFSdirectory d = null;
			try { d = client.getdir(fullpath(p[i])); }
			catch(RequestException e) {
				resp.puts("Error listing "+p[i]+" : "+
					  e.getMessage());
				continue;
				}
			if (p.length > 1)
				resp.puts("Directory of "+p[i]);
			StringBuffer out = new StringBuffer();
			for(int j=0; j<d.size(); j++) {
				out.append(pad(d.find(j).name, 15)+" ");
				if ((j%4) == 3 && j != d.size()-1)
					out.append('\n');
				}
			resp.puts(out.toString());
			}
		}
	}

	// dir
	// List the given directories in long form
	void dir(String p[])
	{
	if (p.length == 0) {
		// List the current directory
		String tmp[] = { client.currentdir };
		dir(tmp);
		}
	else {
		// List the directories given
		for(int i=0; i<p.length; i++) {
			JFSdirectory d = null;
			try { d = client.getdir(fullpath(p[i])); }
			catch(RequestException e) {
				resp.puts("Error listing "+p[i]+" : "+
					  e.getMessage());
				continue;
				}
			resp.puts("Directory of "+p[i]);
			for(int j=0; j<d.size(); j++) {
				JFSfile f = d.find(j);
				String vc = String.valueOf(f.versions.size());
				String line = pad(f.name, 16) + " " +
					      pad(f.type, 16) + " " +
					      pad(String.valueOf(vc), 4);
				for(int k=0; k<f.users.size(); k++) {
					JFSfileuser u = (JFSfileuser)f.users.
								   elementAt(k);
					line += " "+u.name+","+u.rwp;
					}
				resp.puts(line);
				}
			}
		}
	}

	// cd
	// Change to the given directory
	void cd(String p[])
	{
	if (p.length != 1) {
		resp.puts("usage: cd <directory>");
		return;
		}
	try {
		String full = fullpath(p[0]);
		JFSfile f = client.statfile(full);
		if (f.type.equals("jfs/directory"))
			client.currentdir = full;
		else
			resp.puts(p[0]+" is not a directory");
		}
	catch(RequestException e) {
		resp.puts("Error changing directory to "+p[0]+" : "+
			  e.getMessage());
		}
	}

	// pwd
	// Print the current directory and server
	void pwd(String p[])
	{
	resp.puts(client.currentdir+" on "+client.serverhost);
	}

	// open
	// Open the given file using the correct handler, or the handler given
	void open(String p[])
	{
	String path = fullpath(p[0]);
	int ver = version(p[0]);
	if (p.length == 0 || p.length > 2) {
		resp.puts("usage: open <file> [handler]   or <file> [handler]");
		return;
		}
	JFSfile fl = null;
	try { fl = client.statfile(path); }
	catch(RequestException e) {
		resp.puts("Error stating "+p[0]+" : "+e.getMessage());
		return;
		}
	String handler = null;
	if (p.length == 2) {
		// Handler to use is given
		handler = p[1];
		}
	else {
		// Find the correct handler
		try { handler = client.gethandler(fl.type); }
		catch(RequestException e) {
			resp.puts("Error getting handler for "+p[0]+" : "+
				  e.getMessage());
			return;
			}
		if (handler == null) {
			resp.puts("No handler defined for the MIME type "+
				  fl.type);
			return;
			}
		}
	if (ver == 0 && fl.multiversion) {
		// find highest version
		ver = -1;
		for(int i=0; i<fl.versions.size(); i++) {
			int vn = ((JFSversion)fl.versions.elementAt(i)).number;
			if (vn > ver) ver = vn;
			}
		}
	new ProgramWindow(handler, client, path, fl.multiversion ? ver : 0);
	}

	// info
	// Display information about a file or version
	void info(String p[])
	{
	if (p.length != 1) {
		resp.puts("usage: info <file>");
		return;
		}
	String path = fullpath(p[0]);
	int ver = version(p[0]);
	JFSfile f = null;
	try { f = client.statfile(path); }
	catch(RequestException e) {
		resp.puts("Error stating "+p[0]+" : "+e.getMessage());
		return;
		}

	// file details
	resp.puts("Filename: "+path);
	resp.puts("Type:     "+f.type);
	String uline = "";
	for(int i=0; i<f.users.size(); i++) {
		JFSfileuser fu = (JFSfileuser)f.users.elementAt(i);
		uline += fu.name+","+fu.rwp+" ";
		}
	resp.puts("Users:    "+uline);
	if (ver != 0) {
		// Display a specific version
		JFSversion v = null;
		for(int i=0; i<f.versions.size(); i++) {
			JFSversion vv = (JFSversion)f.versions.elementAt(i);
			if (vv.number == ver)
				v = vv;
			}
		if (v == null) {
			resp.puts("Version "+ver+" not found");
			return;
			}
		displayver(v);
		}
	else {
		// Display them all
		resp.puts("Versions: "+(f.multiversion ?
				String.valueOf(f.versions.size()) : "Single"));
		for(int i=0; i<f.versions.size(); i++) {
			JFSversion vv = (JFSversion)f.versions.elementAt(i);
			displayver(vv);
			}
		}
	}

	// displayver
	// Display information about one version of a file
	void displayver(JFSversion v)
	{
	resp.puts("Version:  "+v.number);
	resp.puts("  Size:    "+v.size+" bytes");
	resp.puts("  Creator: "+v.creator);
	resp.puts("  Date:    "+(new Date(v.time)));
	}

	// run
	// Run the given class in a new window, or ask the user to choose one
	void run(String p[])
	{
	if (p.length == 0)
		new ProgramChooser(client);
	else if (p.length == 1)
		new ProgramWindow(p[0], client);
	else {
		resp.puts("usage: run [program]\n");
		return;
		}
	}

	// create
	// Create a new, empty file of some MIME type
	void create(String p[])
	{
	if (p.length != 2) {
		resp.puts("usage: create <file> <type>");
		return;
		}
	String path = fullpath(p[0]);
	int ver = version(p[0]);
	try {
		byte dummy[] = new byte[0];
		client.put(path, ver, dummy, p[1]);
		}
	catch(RequestException e) {
		resp.puts("Failed to create "+p[0]+" : "+e.getMessage());
		}
	}

	// rm
	// Delete the given list of files and directories
	void rm(String p[])
	{
	if (p.length == 0) {
		resp.puts("usage: rm <file> <file> ...");
		return;
		}
	for(int i=0; i<p.length; i++) {
		String path = fullpath(p[i]);
		int ver = version(p[i]);
		try { client.delete(path, ver); }
		catch(RequestException e) {
			resp.puts("Couldn't delete "+p[i]+" : "+e.getMessage());
			return;
			}
		}
	}
}

